/*
	SMALLEST ENCLOSING DISC -SHRINKING ALGORITHM
	Programa que encontra o menor círculo que circunscreve uma dada núvem de pontos
	Implementação do algoritmo (com Tempo Esperado = O(n)) encontrado Toussaint, Computational Geometry, p. 267
	Projeto final do curso de Geometria Computacional, IMPA - 2007
	Marcelo Cicconet, 06 de Junho de 2007
*/

#include <stdio.h>
#include <stdlib.h>
#include <GL/glut.h>
#include <math.h>
#include <time.h>

#define N 10000000
#define PI 3.14159265
#define MIN(a,b) (((a)<(b))?(a):(b))
#define MAX(a,b) (((a)>(b))?(a):(b))

struct	point {
			double x;
			double y;
		};

struct	threepoints {
			struct point v1;
			struct point v2;
			struct point v3;
		};

struct	circle {
			struct point center;
			double radius;
		};

void display(void);
void init(void);
void reshape (int w, int h);
void keyboard(unsigned char key, int x, int y);
struct point makepoint(struct point p1, struct point p2);
struct point makereflpoint(struct point p1, struct point p2);
int incircle(struct point p4, struct point p1, struct point p2, struct point p3);
struct circle cttp(struct threepoints pts); // cttp = circle through three points
int ccw(struct point p1, struct point p2, struct point p3);
int normcomp(struct point p1, struct point p2);
struct point farthest(struct point p[]);
double dist(struct point p1, struct point p2);
struct threepoints minidisc(void);

int width, height;
struct point p[N];
struct circle shsed; // shsed = (shrinking algorithm) smallest enclosing disc
struct threepoints boldthreepoints;
struct threepoints currentcircle;
struct point thethirdpoint;

int main(int argc, char** argv)
{
	int i;
	width = height = 400;

	/* ---------- definindo pontos ---------- */
	/* pontos distribuidos uniformemente em um quadrado */
	for (i = 0; i < N; ++i){
   		p[i].x = 0.2+0.6*((double)rand()/(double)RAND_MAX);
   		p[i].y = 0.2+0.6*((double)rand()/(double)RAND_MAX);
	}
	/* pontos distribuidos uniformemente em um disco */
	/*for (i = 0; i < N; ++i){
   		p[i].x = 0;
   		p[i].y = 0;
   		while ((pow(p[i].x-0.5,2) + pow(p[i].y-0.5,2)) > 0.2){
       		p[i].x = (double)rand()/(double)RAND_MAX;
       		p[i].y = (double)rand()/(double)RAND_MAX;
   		}
	}*/
	/* pontos distribuidos uniformemente na fronteira de um disco */
	/*double aux[N];
    for (i = 0; i < N; ++i){
        aux[i] = 2.0*PI*((double)rand()/(double)RAND_MAX);
        p[i].x = 0.5+0.4*cos(aux[i]);
        p[i].y = 0.5+0.4*sin(aux[i]);
    }*/
	/* pontos distribuidos uniformemente em um triângulo */
    /*for (i = 0; i < N; ++i){
		double pr21, pr32, pr13;
		struct threepoints aux;
		aux.v1.x = 0.2; aux.v1.y = 0.2;
		aux.v2.x = 0.8; aux.v2.y = 0.2;
		aux.v3.x = 0.5; aux.v3.y = 0.8;
        pr21 = 1;
        while (pr21 == 1 || pr32 == 1 || pr13 == 1){
            p[i].x = (double)rand()/(double)RAND_MAX;
            p[i].y = (double)rand()/(double)RAND_MAX;
            pr21 = ccw(aux.v2, aux.v1, p[i]);
            pr32 = ccw(aux.v3, aux.v2, p[i]);
            pr13 = ccw(aux.v1, aux.v3, p[i]);
        }
    }*/
	/* núvem especial de pontos */
	/*for (i = 0; i < floor(N/3.0); ++i){
   		p[i].x = 0.2+0.2*((double)rand()/(double)RAND_MAX);
   		p[i].y = 0.5+0.2*((double)rand()/(double)RAND_MAX);
	}
	for (i = floor(N/3.0); i < floor(2.0*N/3.0); ++i){
   		p[i].x = 0;
   		p[i].y = 0;
   		while ((pow(p[i].x-0.6,2) + pow(p[i].y-0.4,2)) > 0.03){
       		p[i].x = (double)rand()/(double)RAND_MAX;
       		p[i].y = (double)rand()/(double)RAND_MAX;
   		}
	}
	for (i = floor(2.0*N/3.0); i < N; ++i){
		double pr21, pr32, pr13;
		struct threepoints aux;
		aux.v1.x = 0.8; aux.v1.y = 0.5;
		aux.v2.x = 0.8; aux.v2.y = 0.8;
		aux.v3.x = 0.5; aux.v3.y = 0.8;
        pr21 = 1;
        while (pr21 == 1 || pr32 == 1 || pr13 == 1){
            p[i].x = (double)rand()/(double)RAND_MAX;
            p[i].y = (double)rand()/(double)RAND_MAX;
            pr21 = ccw(aux.v2, aux.v1, p[i]);
            pr32 = ccw(aux.v3, aux.v2, p[i]);
            pr13 = ccw(aux.v1, aux.v3, p[i]);
        }
    }*/

	/* iniciando contagem do tempo gasto */
    long int cpu_ini, cpu_fim;
    double tiq_seg;
    tiq_seg = (double)CLOCKS_PER_SEC;
    cpu_ini = clock();

	/* cálculo de três pontos que definem o menor círculo circunscrito */
	currentcircle = minidisc();

	/* finalizando contagem do tempo gasto */
	cpu_fim = clock();
 	printf("Tempo de execução: %g segundos\n", (cpu_fim - cpu_ini)/tiq_seg);

	/* cálculo do centro e do raio do círculo definido pelos pontos encontrados acima */
	shsed = cttp(currentcircle);

	/* visualização */
	boldthreepoints = currentcircle;
	thethirdpoint = makepoint(boldthreepoints.v1, boldthreepoints.v2);
	glutInit(&argc, argv);
    glutInitDisplayMode (GLUT_SINGLE | GLUT_RGB);
    glutInitWindowSize (width, height); 
    glutInitWindowPosition (0, 0);
    glutCreateWindow (argv[0]);
    init ();
    glutDisplayFunc(display); 
    glutReshapeFunc(reshape);
    glutKeyboardFunc(keyboard);
    glutMainLoop();
    return 0;
}

struct threepoints minidisc(void)
{
	int i, side;
	struct point contact1, contact2, reflectedpoint;
	struct circle currentcentrad;

	/* passo 1: encontrando o círculo de centro na origem e passando pelo ponto de maior norma */
	currentcircle.v1 = farthest(p);
	contact1 = currentcircle.v1;
	/* passo 2: reduzindo o círculo acima na direção de currentcircle.v1 */
	if ((currentcircle.v1.x != p[0].x) || (currentcircle.v1.y != p[0].y)){
		currentcircle.v2.x = p[0].x;
		currentcircle.v2.y = p[0].y;
		contact2 = currentcircle.v2;
		reflectedpoint = makereflpoint(currentcircle.v2, currentcircle.v1);
		if (reflectedpoint.x != currentcircle.v2.x || reflectedpoint.y != currentcircle.v2.y){
			currentcircle.v3 = reflectedpoint;
			currentcentrad = cttp(currentcircle);
			side = ccw(currentcentrad.center, contact1, contact2);
		}
		else {
			currentcircle.v3 = makepoint(currentcircle.v1, currentcircle.v2);
			side = 2;
		}
	}
	else {
		currentcircle.v2.x = p[1].x;
		currentcircle.v2.y = p[1].y;
		contact2 = currentcircle.v2;
		reflectedpoint = makereflpoint(currentcircle.v2, currentcircle.v1);
		if (reflectedpoint.x != currentcircle.v2.x || reflectedpoint.y != currentcircle.v2.y){
			currentcircle.v3 = reflectedpoint;
			currentcentrad = cttp(currentcircle);
			side = ccw(currentcentrad.center, contact1, contact2);
		}
		else {
			currentcircle.v3 = makepoint(currentcircle.v1, currentcircle.v2);
			side = 2;
		}
	}
	for (i = 0; i < N; ++i){
		if ((p[i].x != currentcircle.v1.x || p[i].y != currentcircle.v1.y) && (p[i].x != currentcircle.v2.x || p[i].y != currentcircle.v2.y)){
			if (incircle(p[i], currentcircle.v1, currentcircle.v2, currentcircle.v3) == -1){
				currentcircle.v2 = p[i];
				contact2 = currentcircle.v2;
				reflectedpoint = makereflpoint(currentcircle.v2, currentcircle.v1);
				if (reflectedpoint.x != currentcircle.v2.x || reflectedpoint.y != currentcircle.v2.y){
					currentcircle.v3 = reflectedpoint;
					currentcentrad = cttp(currentcircle);
					side = ccw(currentcentrad.center, contact1, contact2);
				}
				else {
					currentcircle.v3 = makepoint(currentcircle.v1, currentcircle.v2);
					side = 2;
				}
			}
			else if (incircle(p[i], currentcircle.v1, currentcircle.v2, currentcircle.v3) == 0){
				if (ccw(currentcentrad.center, contact1, contact2) == side){
					if (dist(p[i], contact1) < dist(p[i], contact2)){
						contact1 = p[i];
					}
					else {
						contact2 = p[i];
					}
				}
			}
		}
	}
	/* passo 3: loop */
	if (ccw(currentcentrad.center, contact1, contact2) != side){
		return currentcircle;
	}
	else {
		while (ccw(currentcentrad.center, contact1, contact2) == side){		
			side = ccw(currentcentrad.center, contact1, contact2);
			currentcircle.v1 = contact1;
			currentcircle.v2 = contact2;
			currentcircle.v3 = makepoint(currentcircle.v1, currentcircle.v2);
			currentcentrad = cttp(currentcircle);
			for (i = 0; i < N; ++i){
				if ((p[i].x != currentcircle.v1.x || p[i].y != currentcircle.v1.y) && (p[i].x != currentcircle.v2.x || p[i].y != currentcircle.v2.y)){
					if (incircle(p[i], currentcircle.v1, currentcircle.v2, currentcircle.v3) == -1){
						currentcircle.v3 = p[i];
						currentcentrad = cttp(currentcircle);
						if (dist(p[i], contact1) < dist(p[i], contact2)){
							contact1 = p[i];
						}
						else {
							contact2 = p[i];
						}
					}
					else if (incircle(p[i], currentcircle.v1, currentcircle.v2, currentcircle.v3) == 0){
						currentcircle.v3 = p[i];
						if (dist(p[i], contact1) < dist(p[i], contact2)){
							contact1 = p[i];
						}
						else {
							contact2 = p[i];
						}						
					}
				}
			}
			if (ccw(currentcentrad.center, contact1, contact2) == 2){
				break;
			}
		}
	return currentcircle;
	}
}

/* círculo passando por três pontos (referência: wolfram mathworld) */
struct circle cttp(struct threepoints pts)
{
	double a, d, e, f, x1, y1, x2, y2, x3, y3;
	struct circle circ;
	x1 = pts.v1.x; y1 = pts.v1.y;
	x2 = pts.v2.x; y2 = pts.v2.y;
	x3 = pts.v3.x; y3 = pts.v3.y;
	a = x1*y2+y1*x3+x2*y3-y2*x3-x1*y3-y1*x2;
	d = -((pow(x1,2)+pow(y1,2))*y2+y1*(pow(x3,2)+pow(y3,2))+(pow(x2,2)+pow(y2,2))*y3-y2*(pow(x3,2)+pow(y3,2))-(pow(x1,2)+pow(y1,2))*y3-y1*(pow(x2,2)+pow(y2,2)));
	e = (pow(x1,2)+pow(y1,2))*x2+x1*(pow(x3,2)+pow(y3,2))+(pow(x2,2)+pow(y2,2))*x3-x2*(pow(x3,2)+pow(y3,2))-(pow(x1,2)+pow(y1,2))*x3-x1*(pow(x2,2)+pow(y2,2));
	f = -((pow(x1,2)+pow(y1,2))*x2*y3+x1*y2*(pow(x3,2)+pow(y3,2))+(pow(x2,2)+pow(y2,2))*x3*y1-x2*y1*(pow(x3,2)+pow(y3,2))-(pow(x1,2)+pow(y1,2))*x3*y2-x1*y3*(pow(x2,2)+pow(y2,2)));
	circ.center.x = -(d/(2.0*a));
	circ.center.y = -(e/(2.0*a));
	circ.radius = sqrt((pow(d,2)+pow(e,2))/(4*pow(a,2)) - f/a);
	return circ;
}

/* incircle retorna -1/0/1 se p4 está fora/sobre/dentro do disco passando por p1, p2, p3 */
int incircle(struct point p4, struct point p1, struct point p2, struct point p3)
{
	/* para p1, p2, e p3 em sentido horário, det > 0 se, e somente se, p4 é ponto interior */
	double det;
	double x1, y1, x2, y2, x3, y3, x4, y4;
	if (ccw(p1, p2, p3) == 1){
		x1 = p1.x; y1 = p1.y;
		x2 = p2.x; y2 = p2.y;
		x3 = p3.x; y3 = p3.y;
		x4 = p4.x; y4 = p4.y;
	}
	else {
		x1 = p3.x; y1 = p3.y;
		x2 = p2.x; y2 = p2.y;
		x3 = p1.x; y3 = p1.y;
		x4 = p4.x; y4 = p4.y;		
	}
	det = (x1-x4)*(y2-y4)*(pow(x3,2)+pow(y3,2)-pow(x4,2)-pow(y4,2))+(y1-y4)*(pow(x2,2)+pow(y2,2)-pow(x4,2)-pow(y4,2))*(x3-x4)+(pow(x1,2)+pow(y1,2)-pow(x4,2)-pow(y4,2))*(x2-x4)*(y3-y4)-(pow(x1,2)+pow(y1,2)-pow(x4,2)-pow(y4,2))*(y2-y4)*(x3-x4)-(pow(x2,2)+pow(y2,2)-pow(x4,2)-pow(y4,2))*(y3-y4)*(x1-x4)-(pow(x3,2)+pow(y3,2)-pow(x4,2)-pow(y4,2))*(y1-y4)*(x2-x4);
	if (det < -0.000000000001){
		return -1;
	}
	else if (det > 0.000000000001){
		return 1;
	}
	else {
		return 0;
	}
}

/* ccw retorna 0/1/2 se p3 está à esquerda/direita/sobre da reta passando por p1, p2 */
int ccw(struct point p1, struct point p2, struct point p3)
{
	double pos, x1, y1, x2, y2, x3, y3;
	x1 = p1.x; y1 = p1.y;
	x2 = p2.x; y2 = p2.y;
	x3 = p3.x; y3 = p3.y;
    pos = ((x1 - x2)*(y3 - y2))-((y1 - y2)*(x3 - x2));
    if (pos > 0.0000000001){
        return 0;
	}
    else if (pos < -0.0000000001){
        return 1;
	}
	else {
		return 2;
	}
}

struct point makepoint(struct point p1, struct point p2)
{
	struct point newpoint;
	newpoint.x = 0.5*(p1.x+p2.x-p1.y+p2.y);
	newpoint.y = 0.5*(p1.y+p2.y+p1.x-p2.x);
	return newpoint;
}

/* Encontra a reflexão de um ponto p1 com respeito à reta definida pelo vetor p2 */
struct point makereflpoint(struct point p1, struct point p2)
{
	struct point reflpoint, normalizedp2, ortonormalizedp2;
	double normp2, c1, c2;
	normp2 = sqrt(pow(p2.x, 2)+pow(p2.y, 2));
	normalizedp2.x = p2.x/normp2;
	normalizedp2.y = p2.y/normp2;
	ortonormalizedp2.x = -normalizedp2.y;
	ortonormalizedp2.y = normalizedp2.x;
	c1 = p1.x*normalizedp2.x+p1.y*normalizedp2.y;
	c2 = p1.x*ortonormalizedp2.x+p1.y*ortonormalizedp2.y;
	reflpoint.x = c1*normalizedp2.x-c2*ortonormalizedp2.x;
	reflpoint.y = c1*normalizedp2.y-c2*ortonormalizedp2.y;
	return reflpoint;
}

double dist(struct point p1, struct point p2)
{
	return (sqrt(pow(p2.x-p1.x, 2)+pow(p2.y-p1.y, 2)));
}

/* encontra o ponto de maior norma */
struct point farthest(struct point p[])
{
	int i;
	struct point thefarthest;
	thefarthest = p[0];
	for (i = 0; i < N-1; ++i){
		if (normcomp(p[i+1], thefarthest) == 1){
			thefarthest = p[i+1];
		}
	}
	return thefarthest;
}

/* retorna 1 se a norma de p1 é maior que a norma de p2, e 0 caso contrário */
int normcomp(struct point p1, struct point p2)
{
	if ((pow(p1.x, 2)+pow(p1.y, 2)) > (pow(p2.x, 2)+pow(p2.y, 2))){
		return 1;
	}
	else {
		return 0;
	}
}

void display(void)
{
    int i;
    glClear(GL_COLOR_BUFFER_BIT);
    glColor3f(0.0, 0.0, 0.0);
	/* desenhando os pontos */
	glPointSize(3.0f);
	glBegin(GL_POINTS);
		for (i = 0; i < N; ++i){
			glVertex2f(p[i].x, p[i].y);
		}
	glEnd();
	/* desenhando o menor círculo circunscrito */
	glColor3f(0.0, 0.0, 1.0);
	glBegin(GL_LINE_LOOP);
		for (i = 0; i < 100; ++i){
			glVertex2f(shsed.center.x+shsed.radius*cos(2.0*PI*(double)i/100.0), shsed.center.y+shsed.radius*sin(2.0*PI*(double)i/100.0));
		}
	glEnd();
	/* realçando os três pontos que definem o menor círculo circunscrito */
	glColor3f(1.0, 0.0, 0.0);
	glBegin(GL_POLYGON);
		for (i = 0; i < 10; ++i){
			glVertex2f(boldthreepoints.v1.x+0.007*cos(2.0*PI*(double)i/10.0), boldthreepoints.v1.y+0.007*sin(2.0*PI*(double)i/10.0));
		}
	glEnd();
	glBegin(GL_POLYGON);
		for (i = 0; i < 10; ++i){
			glVertex2f(boldthreepoints.v2.x+0.007*cos(2.0*PI*(double)i/10.0), boldthreepoints.v2.y+0.007*sin(2.0*PI*(double)i/10.0));
		}
	glEnd();
	if (boldthreepoints.v3.x != thethirdpoint.x || boldthreepoints.v3.y != thethirdpoint.y){
		glBegin(GL_POLYGON);
			for (i = 0; i < 10; ++i){
				glVertex2f(boldthreepoints.v3.x+0.007*cos(2.0*PI*(double)i/10.0), boldthreepoints.v3.y+0.007*sin(2.0*PI*(double)i/10.0));
			}
		glEnd();
	}
    glFlush();
}

void init(void) 
{
    glClearColor (1.0, 1.0, 1.0, 0.0);
    glShadeModel (GL_SMOOTH);
	glEnable(GL_POINT_SMOOTH);
	glHint(GL_POINT_SMOOTH_HINT, GL_NICEST);
}

void reshape (int width, int height)
{
	int min = MIN(width, height);
    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();
	gluOrtho2D(0.0f, 1.0f, 0.0f, 1.0f);
    glMatrixMode(GL_MODELVIEW);
    glLoadIdentity();
    glViewport((width-min)/2, (height-min)/2, min, min);
    glutPostRedisplay();
}

void keyboard(unsigned char key, int x, int y)
{
    switch (key) {
       case 27:
          exit(0);
          break;
    }
}

