/*
	SMALLEST ENCLOSING DISC
	Programa que encontra o menor círculo que circunscreve uma dada núvem de pontos
	Implementação do algoritmo randomizado de Welzl (com Tempo Esperado = O(n)) encontrado em De Berg, p. 85
	Projeto final do curso de Geometria Computacional, IMPA - 2007
	Marcelo Cicconet, 04 de Junho de 2007
*/

#include <stdio.h>
#include <stdlib.h>
#include <GL/glut.h>
#include <math.h>
#include <time.h>

#define N 10000000
#define PI 3.14159265
#define MIN(a,b) (((a)<(b))?(a):(b))
#define MAX(a,b) (((a)>(b))?(a):(b))

struct	point {
			double x;
			double y;
		};

struct	threepoints {
			struct point v1;
			struct point v2;
			struct point v3;
		};

struct	circle {
			struct point center;
			double radius;
		};

void display(void);
void init(void);
void reshape (int w, int h);
void keyboard(unsigned char key, int x, int y);
struct threepoints minidiscwith2points(int n, struct point q1, struct point q2);
struct threepoints minidiscwithpoint(int n, struct point q);
struct threepoints minidisc(void);
struct point makepoint(struct point p1, struct point p2);
int incircle(struct point p4, struct point p1, struct point p2, struct point p3);
struct circle cttp(struct threepoints pts); // cttp = circle through three points
int ccw(struct point p1, struct point p2, struct point p3);

int width, height;
struct point p[N];
struct circle sed; // sed = smallest enclosing disc
struct threepoints boldthreepoints;
struct point thethirdpoint;

int main(int argc, char** argv)
{
    int    i;
	struct threepoints currentcircle;
	width = height = 400;

	/* ---------- definindo pontos DETERMINISTICAMENTE ---------- */
	/* pontos distribuidos em uma curva senoidal */
	/*for (i = 0; i < N; ++i){
		p[i].x = 0.2+0.6*(double)i/(double)N;
		p[i].y = 0.5+0.3*sin(100.0*p[i].x);
	}*/
	/* pontos distribuidos em uma espiral */
	/*for (i = 0; i < N; ++i){
        p[i].x = 0.5+0.4*((double)i/(double)N)*cos(20.0*PI*(double)i/(double)N);
        p[i].y = 0.5+0.4*((double)i/(double)N)*sin(20.0*PI*(double)i/(double)N);
    }*/
	/* pontos distribuidos em uma reta */
	/*for (i = 0; i < N; ++i){
		p[i].x = 0.2+0.6*((double)i/(double)N);
		p[i].y = 0.2+0.6*p[i].x;
	}*/
	/* ---------- definindo pontos ALEATORIAMENTE ---------- */
	/* pontos distribuidos uniformemente em um quadrado */
	/*for (i = 0; i < N; ++i){
   		p[i].x = 0.2+0.6*((double)rand()/(double)RAND_MAX);
   		p[i].y = 0.2+0.6*((double)rand()/(double)RAND_MAX);
	}*/
	/* pontos distribuidos uniformemente em um disco */
	/*for (i = 0; i < N; ++i){
   		p[i].x = 0;
   		p[i].y = 0;
   		while ((pow(p[i].x-0.5,2) + pow(p[i].y-0.5,2)) > 0.2){
       		p[i].x = (double)rand()/(double)RAND_MAX;
       		p[i].y = (double)rand()/(double)RAND_MAX;
   		}
	}*/
	/* pontos distribuidos uniformemente na fronteira de um disco */
	/*double aux[N];
    for (i = 0; i < N; ++i){
        aux[i] = 2.0*PI*((double)rand()/(double)RAND_MAX);
        p[i].x = 0.5+0.4*cos(aux[i]);
        p[i].y = 0.5+0.4*sin(aux[i]);
    }*/
	/* pontos distribuidos uniformemente em um triângulo */
    for (i = 0; i < N; ++i){
		double pr21, pr32, pr13;
		struct threepoints aux;
		aux.v1.x = 0.2; aux.v1.y = 0.2;
		aux.v2.x = 0.8; aux.v2.y = 0.2;
		aux.v3.x = 0.5; aux.v3.y = 0.8;
        pr21 = 1;
        while (pr21 == 1 || pr32 == 1 || pr13 == 1){
            p[i].x = (double)rand()/(double)RAND_MAX;
            p[i].y = (double)rand()/(double)RAND_MAX;
            pr21 = ccw(aux.v2, aux.v1, p[i]);
            pr32 = ccw(aux.v3, aux.v2, p[i]);
            pr13 = ccw(aux.v1, aux.v3, p[i]);
        }
    }
	/* núvem especial de pontos */
	/*for (i = 0; i < floor(N/3.0); ++i){
   		p[i].x = 0.2+0.2*((double)rand()/(double)RAND_MAX);
   		p[i].y = 0.5+0.2*((double)rand()/(double)RAND_MAX);
	}
	for (i = floor(N/3.0); i < floor(2.0*N/3.0); ++i){
   		p[i].x = 0;
   		p[i].y = 0;
   		while ((pow(p[i].x-0.6,2) + pow(p[i].y-0.4,2)) > 0.03){
       		p[i].x = (double)rand()/(double)RAND_MAX;
       		p[i].y = (double)rand()/(double)RAND_MAX;
   		}
	}
	for (i = floor(2.0*N/3.0); i < N; ++i){
		double pr21, pr32, pr13;
		struct threepoints aux;
		aux.v1.x = 0.8; aux.v1.y = 0.5;
		aux.v2.x = 0.8; aux.v2.y = 0.8;
		aux.v3.x = 0.5; aux.v3.y = 0.8;
        pr21 = 1;
        while (pr21 == 1 || pr32 == 1 || pr13 == 1){
            p[i].x = (double)rand()/(double)RAND_MAX;
            p[i].y = (double)rand()/(double)RAND_MAX;
            pr21 = ccw(aux.v2, aux.v1, p[i]);
            pr32 = ccw(aux.v3, aux.v2, p[i]);
            pr13 = ccw(aux.v1, aux.v3, p[i]);
        }
    }*/

	/* iniciando contagem do tempo gasto */
    long int cpu_ini, cpu_fim;
    double tiq_seg;
    tiq_seg = (double)CLOCKS_PER_SEC;
    cpu_ini = clock();

	/* cálculo de três pontos que definem o menor círculo circunscrito */
	currentcircle = minidisc();

	/* finalizando contagem do tempo gasto */
	cpu_fim = clock();
    printf("Tempo de execução: %g segundos\n", (cpu_fim - cpu_ini)/tiq_seg);

	/* cálculo do centro e do raio do círculo definido pelos pontos encontrados acima */
	sed = cttp(currentcircle);
	
	/* visualização */
	boldthreepoints = currentcircle;
	thethirdpoint = makepoint(boldthreepoints.v1, boldthreepoints.v2);
	glutInit(&argc, argv);
    glutInitDisplayMode (GLUT_SINGLE | GLUT_RGB);
    glutInitWindowSize (width, height); 
    glutInitWindowPosition (0, 0);
    glutCreateWindow (argv[0]);
    init ();
    glutDisplayFunc(display); 
    glutReshapeFunc(reshape);
    glutKeyboardFunc(keyboard);
    glutMainLoop();
    return 0;	
}

struct threepoints minidisc(void)
{
	int i;
	struct threepoints currentcircle;
	currentcircle.v1 = p[0];
	currentcircle.v2 = p[1];
	currentcircle.v3 = makepoint(p[0], p[1]);
	for (i = 2; i < N; ++i){
		if (incircle(p[i], currentcircle.v1, currentcircle.v2, currentcircle.v3) == 0){
			currentcircle = minidiscwithpoint(i, p[i]);
		}
	}
	return currentcircle;
}

struct threepoints minidiscwithpoint(int n, struct point q)
{
	int i;
	struct threepoints currentcircle;
	currentcircle.v1 = q;
	currentcircle.v2 = p[0];
	currentcircle.v3 = makepoint(q, p[0]);
	for (i = 1; i < n; ++i){
		if (incircle(p[i], currentcircle.v1, currentcircle.v2, currentcircle.v3) == 0){
			currentcircle = minidiscwith2points(i, q, p[i]);
		}
	}
	return currentcircle;
}

struct threepoints minidiscwith2points(int n, struct point q1, struct point q2)
{
	int i;
	struct threepoints currentcircle;
	currentcircle.v1 = q1;
	currentcircle.v2 = q2;
	currentcircle.v3 = makepoint(q1, q2);
	for (i = 0; i < n; ++i){
		if (incircle(p[i], currentcircle.v1, currentcircle.v2, currentcircle.v3) == 0){
			currentcircle.v3 = p[i];
		}
	}
	return currentcircle;
}

/* círculo passando por três pontos (referência: wolfram mathworld) */
struct circle cttp(struct threepoints pts)
{
	double a, d, e, f, x1, y1, x2, y2, x3, y3;
	struct circle circ;
	x1 = pts.v1.x; y1 = pts.v1.y;
	x2 = pts.v2.x; y2 = pts.v2.y;
	x3 = pts.v3.x; y3 = pts.v3.y;
	a = x1*y2+y1*x3+x2*y3-y2*x3-x1*y3-y1*x2;
	d = -((pow(x1,2)+pow(y1,2))*y2+y1*(pow(x3,2)+pow(y3,2))+(pow(x2,2)+pow(y2,2))*y3-y2*(pow(x3,2)+pow(y3,2))-(pow(x1,2)+pow(y1,2))*y3-y1*(pow(x2,2)+pow(y2,2)));
	e = (pow(x1,2)+pow(y1,2))*x2+x1*(pow(x3,2)+pow(y3,2))+(pow(x2,2)+pow(y2,2))*x3-x2*(pow(x3,2)+pow(y3,2))-(pow(x1,2)+pow(y1,2))*x3-x1*(pow(x2,2)+pow(y2,2));
	f = -((pow(x1,2)+pow(y1,2))*x2*y3+x1*y2*(pow(x3,2)+pow(y3,2))+(pow(x2,2)+pow(y2,2))*x3*y1-x2*y1*(pow(x3,2)+pow(y3,2))-(pow(x1,2)+pow(y1,2))*x3*y2-x1*y3*(pow(x2,2)+pow(y2,2)));
	circ.center.x = -(d/(2.0*a));
	circ.center.y = -(e/(2.0*a));
	circ.radius = sqrt((pow(d,2)+pow(e,2))/(4*pow(a,2)) - f/a);
	return circ;
}

/* incircle retorna 0 se p4 está fora do disco passando por p1, p2, p3 e retorna 1 caso contrário */
int incircle(struct point p4, struct point p1, struct point p2, struct point p3)
{
	/* para p1, p2, e p3 em sentido horário, det > 0 se, e somente se, p4 é ponto interior */
	double det;
	double x1, y1, x2, y2, x3, y3, x4, y4;
	if (ccw(p1, p2, p3) == 1){
		x1 = p1.x; y1 = p1.y;
		x2 = p2.x; y2 = p2.y;
		x3 = p3.x; y3 = p3.y;
		x4 = p4.x; y4 = p4.y;
	}
	else {
		x1 = p3.x; y1 = p3.y;
		x2 = p2.x; y2 = p2.y;
		x3 = p1.x; y3 = p1.y;
		x4 = p4.x; y4 = p4.y;		
	}
	det = (x1-x4)*(y2-y4)*(pow(x3,2)+pow(y3,2)-pow(x4,2)-pow(y4,2))+(y1-y4)*(pow(x2,2)+pow(y2,2)-pow(x4,2)-pow(y4,2))*(x3-x4)+(pow(x1,2)+pow(y1,2)-pow(x4,2)-pow(y4,2))*(x2-x4)*(y3-y4)-(pow(x1,2)+pow(y1,2)-pow(x4,2)-pow(y4,2))*(y2-y4)*(x3-x4)-(pow(x2,2)+pow(y2,2)-pow(x4,2)-pow(y4,2))*(y3-y4)*(x1-x4)-(pow(x3,2)+pow(y3,2)-pow(x4,2)-pow(y4,2))*(y1-y4)*(x2-x4);
	if (det < 0.0){
		return 0;
	}
	else {
		return 1;
	}
}

/* ccw retorna 0 (resp., 1) se p3 está à esquerda (respec., direita) da reta passando por p1, p2 */
int ccw(struct point p1, struct point p2, struct point p3)
{
	double pos, x1, y1, x2, y2, x3, y3;
	x1 = p1.x; y1 = p1.y;
	x2 = p2.x; y2 = p2.y;
	x3 = p3.x; y3 = p3.y;
    pos = ((x1 - x2)*(y3 - y2))-((y1 - y2)*(x3 - x2));
    if (pos > 0.0){
        return 0;
	}
    else if (pos < 0.0){
        return 1;
	}
}

struct point makepoint(struct point p1, struct point p2)
{
	struct point newpoint;
	newpoint.x = 0.5*(p1.x+p2.x-p1.y+p2.y);
	newpoint.y = 0.5*(p1.y+p2.y+p1.x-p2.x);
	return newpoint;
}

void display(void)
{
    int i;
    glClear(GL_COLOR_BUFFER_BIT);
    glColor3f(0.0, 0.0, 0.0);
	/* desenhando os pontos */
	glPointSize(3.0f);
	glBegin(GL_POINTS);
		for (i = 0; i < N; ++i){
			glVertex2f(p[i].x, p[i].y);
		}
	glEnd();
	/* desenhando o menor círculo circunscrito */
	glColor3f(0.0, 0.0, 1.0);
	glBegin(GL_LINE_LOOP);
		for (i = 0; i < 100; ++i){
			glVertex2f(sed.center.x+sed.radius*cos(2.0*PI*(double)i/100.0), sed.center.y+sed.radius*sin(2.0*PI*(double)i/100.0));
		}
	glEnd();
	/* realçando os três pontos que definem o menor círculo circunscrito */
	glColor3f(1.0, 0.0, 0.0);
	glBegin(GL_POLYGON);
		for (i = 0; i < 10; ++i){
			glVertex2f(boldthreepoints.v1.x+0.007*cos(2.0*PI*(double)i/10.0), boldthreepoints.v1.y+0.007*sin(2.0*PI*(double)i/10.0));
		}
	glEnd();
	glBegin(GL_POLYGON);
		for (i = 0; i < 10; ++i){
			glVertex2f(boldthreepoints.v2.x+0.007*cos(2.0*PI*(double)i/10.0), boldthreepoints.v2.y+0.007*sin(2.0*PI*(double)i/10.0));
		}
	glEnd();
	if (boldthreepoints.v3.x != thethirdpoint.x || boldthreepoints.v3.y != thethirdpoint.y){
		glBegin(GL_POLYGON);
			for (i = 0; i < 10; ++i){
				glVertex2f(boldthreepoints.v3.x+0.007*cos(2.0*PI*(double)i/10.0), boldthreepoints.v3.y+0.007*sin(2.0*PI*(double)i/10.0));
			}
		glEnd();
	}
    glFlush();
}

void init(void) 
{
    glClearColor (1.0, 1.0, 1.0, 0.0);
    glShadeModel (GL_SMOOTH);
	glEnable(GL_POINT_SMOOTH);
	glHint(GL_POINT_SMOOTH_HINT, GL_NICEST);
}

void reshape (int width, int height)
{
	int min = MIN(width, height);
    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();
	gluOrtho2D(0.0f, 1.0f, 0.0f, 1.0f);
    glMatrixMode(GL_MODELVIEW);
    glLoadIdentity();
    glViewport((width-min)/2, (height-min)/2, min, min);
    glutPostRedisplay();
}

void keyboard(unsigned char key, int x, int y)
{
    switch (key) {
       case 27:
          exit(0);
          break;
    }
}

